/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
/* all functions here assume that vectors are allocated
you can pass a vector in the 'source' arguments and at the 'result' in all
functions */

#include "mathlib.h"
#include <math.h>
VECTOR_UL prv;

void initV ( VECTOR_UL *vect, double x, double y, double z ) {
	vect->x = x;
	vect->y = y;
	vect->z = z;
}

void zeroV ( VECTOR_UL *vect ) {
	/* memset( vect, 0, sizeof(VECTOR_UL)) is not portable */
	vect->x = vect->y = vect->z = 0.0;
}

double normV ( VECTOR_UL *vect ) {
	return sqrt( vect->x*vect->x + vect->y*vect->y + vect->z*vect->z );
}

void copyV ( VECTOR_UL *target, const VECTOR_UL *src ) {
	memcpy( target, src, sizeof(VECTOR_UL) );
}

void copyPtoV ( VECTOR_UL *target, const POINT_UL *src ) {
	memcpy( target, src->vect, sizeof(VECTOR_UL) );
}

void copyVtoP ( POINT_UL *target, const VECTOR_UL *src ) {
	memcpy( target->vect, src, sizeof(VECTOR_UL) );
}

void addV ( VECTOR_UL *vect1, VECTOR_UL *vect2, VECTOR_UL *result ) {
	result->x = vect1->x + vect2->x;
	result->y = vect1->y + vect2->y;
	result->z = vect1->z + vect2->z;		
}

void subV ( VECTOR_UL *vect1, VECTOR_UL *vect2, VECTOR_UL *result ) {
	result->x = vect1->x - vect2->x;
	result->y = vect1->y - vect2->y;
	result->z = vect1->z - vect2->z;		
}

void scalarMV ( VECTOR_UL *vect, const double scalar, VECTOR_UL *result ) {
	result->x = vect->x * scalar;
	result->y = vect->y * scalar;
	result->z = vect->z * scalar;
}

void normalizeV ( VECTOR_UL *vect ) {
	double mod = normV( vect );
	if ( mod != 0 )
		scalarMV( vect, (1/mod), vect );
}

double dotprodV ( VECTOR_UL *vect1, VECTOR_UL *vect2 ) {
	return sqrt(vect1->x*vect2->x + vect1->y*vect2->y + vect1->z*vect2->z);
}

void crossprodV ( VECTOR_UL *vect1, VECTOR_UL *vect2, VECTOR_UL *result ) {
	if ( vect1 != result && vect2 != result ) {
		result->x = vect1->y*vect2->z - vect1->z*vect2->y;
		result->y = vect1->z*vect2->x - vect1->x*vect2->z;
		result->z = vect1->x*vect2->y - vect1->y*vect2->x;
		return;
	}
	else {
		prv.x = vect1->y*vect2->z - vect1->z*vect2->y;
		prv.y = vect1->z*vect2->x - vect1->x*vect2->z;
		prv.z = vect1->x*vect2->y - vect1->y*vect2->x;
		result->x = prv.x;
		result->y = prv.y;
		result->z = prv.z;
	}
}

double dist ( VECTOR_UL *v1, VECTOR_UL *v2 ) {
	VECTOR_UL v;
	subV( v2, v1, &v );
	return( normV( &v ) );
}

/* squared distance, faster */
double sqdist ( VECTOR_UL *v1, VECTOR_UL *v2 ) {
	VECTOR_UL v;
	subV( v2, v1, &v );
	return( v.x*v.x + v.y*v.y + v.z*v.z );
}

/* assumes b != c != a */
void multM3 ( BASISCHANGE_UL a, BASISCHANGE_UL b, BASISCHANGE_UL c ) {
	int i, j, k;
	for ( i = 0; i < 3; i++ )
		for ( j = 0; j < 3; j++ )
			for ( k = 0; k < 3; k++ )
				c[i][j] = a[i][k]+b[k][j];

}

/* if bases are orthonormal, then the basis change matrix is orthogonal,
because it's only a rotation. From this follows that the inverse is the
transpose. */
void invertbasis( BASISCHANGE_UL orig, BASISCHANGE_UL inv ) {
	int i, j;
	BASISCHANGE_UL b;
	if ( &orig[0][0] != &inv[0][0] )
		for ( i = 0; i < 3; i++ )
			for ( j = 0; j < 3; j++ )
				inv[j][i] = orig[i][j];
	else {
		for ( i = 0; i < 3; i++ )
			for ( j = 0; j < 3; j++ )
				b[j][i] = orig[i][j];
		for ( i = 0; i < 3; i++ )
			for ( j = 0; j < 3; j++ )
				inv[i][j] = b[i][j];
	}
} 

void changebasis ( VECTOR_UL *v, BASISCHANGE_UL b, VECTOR_UL *r ) {
	if ( v != r ) {
		r->x = b[0][0] * v->x + b[0][1] * v->y + b[0][2] * v->z;
		r->y = b[1][0] * v->x + b[1][1] * v->y + b[1][2] * v->z;
		r->z = b[2][0] * v->x + b[2][1] * v->y + b[2][2] * v->z;
		return;
	}
	else {
		prv.x = b[0][0] * v->x + b[0][1] * v->y + b[0][2] * v->z;
		prv.y = b[1][0] * v->x + b[1][1] * v->y + b[1][2] * v->z;
		prv.z = b[2][0] * v->x + b[2][1] * v->y + b[2][2] * v->z;
		copyV( r, &prv );
	}
}

/* given a vector and an angle, this returns the basis change matrix from
canonical to it */
/***********SHOULD BE TESTED */
void basisVA ( VECTOR_UL *v, double angle, BASISCHANGE_UL m ) {
	double a, b, d;
	VECTOR_UL fy;
	QUATERNION_UL rot, invrot;
	
	/* new basis (let's call f) x component is v, normalized */
	copyV( &prv, v );
	normalizeV( &prv );
	m[0][0] = prv.x;	m[0][1] = prv.y;	m[0][2] = prv.z;
	d = sqrt(prv.x*prv.x + prv.y*prv.y );
	a = prv.y / d;
	b = prv.x / d;
	if ( prv.x > 0 && prv.y > 0 ) {
		a *= -1;
	}
	if ( prv.x > 0 && prv.y < 0 ) {
		a *= -1;
		b *= -1;
	} 
	if ( prv.x < 0 && prv.y > 0 ) {
		b *= -1;
	}
	/* this is fy */
	fy.x = a;	      /* these values */
	fy.y = b;	      /* are for */
	fy.z = 0;	      /* angle = 0 */
	
	/* now rotate */
	if ( angle ) {
		initQuatVA( &rot, angle, &prv );
		conjugateQ( &rot, &invrot );
		qvq( &rot, &fy, &invrot, &fy );
	}
	normalizeV( &fy );
	m[1][0] = fy.x;		m[1][1] = fy.y;		m[1][2] = fy.z;
	
	crossprodV( &prv, &fy, &prv );
	normalizeV( &prv );	/* it should be normalized, but it doesn't hurt */
	m[2][0] = prv.x;	m[2][1] = prv.y;	m[2][2] = prv.z;
}
