/* sound.c */
/* Placed in public domain by Steve Kirkendall.  Use at your own risk */

/* This file contains the soundfind() function, which reads sounds from a file
 * into an in-memory structure.  The sounds are never freed.
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "unixsnd.h"
#include "data.h"
#include "ulaw.h"
#define QTY(array)	(sizeof(array) / sizeof(*(array)))

/* Just in case <stdio.h> doesn't define it... */
#ifndef SEEK_SET
# define SEEK_SET	0
#endif

/* To avoid dependency on byte ordering, any multi-byte values in sound headers
 * are given as arrays of unsigned chars.  Macros are used to produce 32-bit
 * values from these.
 */
typedef	unsigned char	big4_t[4];
#define getbig4(f)	 (((f)[0]<<24) | ((f)[1]<<16) | ((f)[2]<<8) | (f)[3])
typedef unsigned char	ltl4_t[4];
#define getltl4(f)	 (((f)[3]<<24) | ((f)[2]<<16) | ((f)[1]<<8) | (f)[0])
typedef union
{
	struct
	{
		char	dotsnd[4];	/* ".snd" */
		big4_t	dataoffset;	/* where the sound data starts */
		big4_t	bytes;		/* number of bytes of sound data */
		char	reserved[3];
		char	encoding;	/* 1=mulaw, 2=u8, 3=u16be, 23=adpcm */
		big4_t	samplefreq;	/* sample frequency */
		char	reserved2[3];
		char	channels;	/* 1=mono, 2=stereo */
	} snd;

	struct
	{
		char	riff[4];	/* "RIFF" */
		ltl4_t	nextdelta;	/* offset to EOF or next segment */
		char	wave[4];	/* "WAVE" */
		char	fmt[4];		/* "fmt " */
		char	reserved1[6];
		char	channels;	/* 1=mono, 2=stereo */
		char	reserved2[1];
		ltl4_t	samplefreq;	/* sample frequency */
		char	reserved3[4];
		char	encoding[4];	/* I *hope* this is the encoding */
		char	data[4];	/* "data" */
		ltl4_t	bytes;		/* number of bytes of sound data */
	} wav;

	unsigned char	byte[64];
} hdr_t;


char *path;

/* some built-in sounds */
static sound_t tone = {NULL,  "tone", 16000, 0, QTY(tonedata), tonedata};
static sound_t buzz = {&tone, "buzz", 16000, 0, QTY(buzzdata), buzzdata};
static sound_t hiss = {&buzz, "hiss", 16000, 0, QTY(hissdata), hissdata};
static sound_t bang = {&hiss, "bang", 16000, 0, QTY(hissdata), bangdata};

/* a list of loaded sounds */
static sound_t *loaded = &bang;


/* The following functions all read a single sample from a file, and convert
 * it to 16-bit signed format.
 */

/* get a signed 16-bit little-endian sample */
static short get_s16le(FILE *fp)
{
	int	lobyte = getc(fp);
	return lobyte | (getc(fp) << 8);
}

#if 0
/* get a signed 16-bit big-endian sample */
static short get_s16be(FILE *fp)
{
	int	hibyte = getc(fp);
	return (short)(hibyte << 8) | getc(fp);
}

/* get an unsigned 16-bit little-endian sample */
static short get_u16le(FILE *fp)
{
	int	lobyte = getc(fp);
	return (lobyte | (getc(fp) << 8)) ^ 0x8000;
}
#endif

/* get an unsigned 16-bit big-endian sample */
static short get_u16be(FILE *fp)
{
	int	hibyte = getc(fp);
	return ((hibyte << 8) | getc(fp)) ^ 0x8000;
}

#if 0
/* get a signed 8-bit sample */
static short get_s8(FILE *fp)
{
	return getc(fp) << 8;
}
#endif

/* get an unsigned 8-bit sample */
static short get_u8(FILE *fp)
{
	return (getc(fp) << 8) ^ 0x8000;
}

/* get a mu-Law sample */
static short get_ulaw(FILE *fp)
{
	return ulaw2linear[getc(fp) & 0xff];
}


/* Search for a given file.  "basename" normally contains no directory name
 * or filename extension, although it may have either of those.
 */
static FILE *searchpath(char *basename)
{
	char	fullname[300];
	char	*cp;
	char	*home;
	int	len, i;
	FILE	*fp;

	/* If name contains any slashes, then assume it is an absolute or
	 * relative file name.  Don't search the path.
	 */
	if (*basename == '/')
	{
		/* use the name as-is */
		strcpy(fullname, basename);
		len = strlen(basename);

		/* try to open the file with no extension, or ".au" or ".wav" */
		fp = fopen(fullname, "r");
		if (!fp)
		{
			strcpy(&fullname[len], ".au");
			fp = fopen(fullname, "r");
		}
		if (!fp)
		{
			strcpy(&fullname[len], ".wav");
			fp = fopen(fullname, "r");
		}
	}
	else /* search through each directory name in the sound path */
	{
		for (cp = path, fp = NULL;
		     !fp && *cp;
		     cp = (*cp ? cp + 1 : cp))
		{
			/* copy the directory into fullname */
			len = 0;
			if (*cp == '~')
			{
				home = getenv("HOME");
				if (home)
				{
					strcpy(fullname, home);
					len = strlen(home);
				}
			}
			while (*cp && *cp != ':')
				fullname[len++] = *cp++;
			if (len == 0)
				fullname[len++] = '.';

			/* append a '/' */
			fullname[len++] = '/';

			/* append the basename */
			strcpy(&fullname[len], basename);
			len += strlen(basename);

			/* try to open the file with no extension, or ".au"
			 * or ".wav"
			 */
			fp = fopen(fullname, "r");
			if (!fp)
			{
				strcpy(&fullname[len], ".au");
				fp = fopen(fullname, "r");
			}
			if (!fp)
			{
				strcpy(&fullname[len], ".wav");
				fp = fopen(fullname, "r");
			}
		}
	}
	if (verbose)
		if (fp)
			printf("loading sound from file %s\n", fullname);
		else
			printf("could not find sound file for \"%s\"\n", basename);
	return fp;
}


/* Read a header, and stuff the parameters (sample frequency, stereo/mono,
 * number of samples) into the sound struct.  Leave the file pointer at the
 * start of the data.  Return a pointer to a function which will get a single
 * sample.  If there is an error such as an unsupported encoding, then give
 * an error message and return NULL.
 *
 * The declaration is unbelievably obscure.  This function takes two arguments:
 * a FILE* and a sound_t*.  It returns a pointer to a function which will take
 * a single FILE* argument and return a short.
 */
static short (*readheader(FILE *fp, sound_t *sound))(FILE *fp)
{
	hdr_t	hdr;
	int	i;

	/* Attempt to read the header.  If the file is shorter than a hdr_t,
	 * then assume it has no header.
	 */
	if (fread(&hdr, sizeof hdr, 1, fp) < 1)
		goto NoHeader;

	/* Determine which type of header we have. */
	if (!memcmp(hdr.snd.dotsnd, ".snd", 4))
	{
SndHeader:	/* we have a Sun ".snd" header, probably in a *.au file. */
		sound->frequency = getbig4(hdr.snd.samplefreq);
		sound->stereo = (hdr.snd.channels == 2 ? 1 : 0);
		sound->nsamples = getbig4(hdr.snd.bytes);
		if (hdr.snd.encoding == 3)
			sound->nsamples >>= 1;
		fseek(fp, getbig4(hdr.snd.dataoffset), SEEK_SET);
		if (verbose)
			printf(".snd encoding #%d, %dHz, %s, offset=%d, nsamples=%d\n",
				hdr.snd.encoding,
				sound->frequency,
				sound->stereo ? "stereo" : "mono",
				getbig4(hdr.snd.dataoffset),
				sound->nsamples);
		switch (hdr.snd.encoding)
		{
		  case 1:	return get_ulaw;
		  case 2:	return get_u8;
		  case 3:	return get_u16be;
		}
	}
	else if (!memcmp(hdr.wav.riff, "RIFF", 4))
	{
WavHeader:	/* we have a Microsoft "WAV" header. */
		sound->frequency = getltl4(hdr.wav.samplefreq);
		sound->stereo = (hdr.wav.channels == 2) ? 1 : 0;
		if (!memcmp(hdr.wav.data, "data", 4))
		{
			sound->nsamples = getltl4(hdr.wav.bytes);
			fseek(fp, (long)sizeof hdr.wav, SEEK_SET);
		}
		else
		{
			for (i = 0; i < sizeof hdr - 4; i++)
				if (!memcmp(&hdr.byte[i], "data", 4))
					break;
			sound->nsamples = getltl4(&hdr.byte[i + 4]);
			fseek(fp, (long)(i + 8), SEEK_SET);
		}
		if (hdr.wav.encoding[0] / hdr.wav.channels == 2)
			sound->nsamples >>= 1;
		if (verbose)
			printf("WAV encoding #%d, %dHz, %s, offset=%ld, nsamples=%d\n",
				hdr.wav.encoding[0],
				sound->frequency,
				sound->stereo ? "stereo" : "mono",
				ftell(fp),
				sound->nsamples);
		switch (hdr.wav.encoding[0] / hdr.wav.channels)
		{
		  case 1:	return get_u8;
		  case 2:	return get_s16le;
		}
	}
	else
	{
NoHeader:	/* no header recognized.  Assume headerless *.au file */
		sound->frequency = 8000;
		sound->stereo = 0;
		sound->nsamples = 0;
		if (verbose)
			printf("headerless encoding u-Law, 8000Hz, mono, offset=0, nsamples=unknown\n");
		fseek(fp, 0L, SEEK_SET);
		return get_ulaw;
	}

	fprintf(stderr, "unixsnd: unsupported encoding\n");
	return NULL;
}

sound_t *soundfind(char *basename)
{
	FILE	*fp;
	sound_t	*ret;
	short	(*get)(FILE *fp);
	short	sample;
	int	allocated;
	short	*newmem;


	/* first check the already-loaded sounds */
	for (ret = loaded; ret; ret = ret->next)
	{
		if (!strcmp(ret->name, basename))
			return ret;
	}

	/* Search for the sound file.  If not found then return NULL */
	fp = searchpath(basename);
	if (!fp)
		return NULL;

	/* allocate a struct for storing info about this sound */
	ret = (sound_t *)malloc(sizeof(sound_t));
	ret->next = loaded;
	loaded = ret;
	ret->name = strdup(basename);
	ret->frequency = 8000;

	/* read the header */
	get = readheader(fp, ret);
	if (!get)
	{
		free(ret);
		return NULL;
	}

	/* read the data, converting samples to 16-bit signed values */
	if (ret->nsamples == 0)
	{
		/* don't know how many samples we have */
		allocated = 0;
		while (sample = (*get)(fp), !feof(fp))
		{
			if (ret->nsamples >= allocated)
			{
				allocated += 512;
				newmem = (short *)malloc(allocated * sizeof(short));
				if (ret->nsamples > 0)
				{
					memcpy(newmem, ret->sample,
						ret->nsamples * sizeof(short));
					free(ret->sample);
				}
				ret->sample = newmem;
			}
			ret->sample[ret->nsamples++] = sample;
		}
	}
	else
	{
		/* we know exactly how many samples to read */
		ret->sample = (short *)malloc(ret->nsamples * sizeof(short));
		for (allocated = 0; allocated < ret->nsamples; allocated++)
		{
			ret->sample[allocated] = (*get)(fp);
		}
	}

	/* close the sound file */
	fclose(fp);

	/* return the sound */
	return ret;
}


/* Add one or more directories to the sound path */
void soundpath(char *addpath)
{
	char	*newpath;

	/* was there a previous path? */
	if (!path)
	{
		/* No, just use the new path */
		path = strdup(addpath);
	}
	else
	{
		/* Yes, combine them */
		newpath = (char *)malloc(strlen(addpath) + strlen(path) + 2);
		strcpy(newpath, addpath);
		strcat(newpath, ":");
		strcat(newpath, path);
		free(path);
		path = newpath;
	}
}
